#include <string>
#include "glm/ext.hpp"
#include "IDriver.h"
#include "ICore.h"
#include "IScene.h"
#include "INode.h"
#include "IMesh.h"
#include "IGpuProgram.h"
#include "INodeDelegate.h"
#include "SRenderState.h"

const std::string simpleVs =
        "attribute vec4 POSITION;\n"
        "attribute vec3 NORMAL;\n"
        "attribute vec2 TEXCOORD;\n"
        "uniform mat4 PROJ_MATRIX;\n"
        "uniform mat4 VIEW_MATRIX;\n"
        "uniform mat4 MODEL_MATRIX;\n"
        "uniform mat3 NORMAL_MATRIX;\n"
        "uniform vec3 LIGHTS_POSITION;\n"
        "varying vec2 texcoord;\n"
        "varying vec3 vVaryingLightDir;\n"
        "varying vec3 vVaryingNormal;\n"
        "void main(){\n"
        "    vVaryingNormal = NORMAL;\n"
        "    vec4 vPosition4 = MODEL_MATRIX * POSITION;\n"
        "    vec3 vPosition3 = vPosition4.xyz / vPosition4.w;\n"
        "    vVaryingLightDir = -normalize(LIGHTS_POSITION - vPosition3);\n"
        "    gl_Position = PROJ_MATRIX * VIEW_MATRIX * MODEL_MATRIX * POSITION;\n"
        "    texcoord = TEXCOORD;\n"
        "}";

const std::string simpleFs =
        "uniform sampler2D TEXTURE0;\n"
        "uniform sampler2D NORMALMAP;\n"
        "uniform vec4 LIGHTS_AMBIENT;\n"
        "uniform vec4 LIGHTS_DIFFUSE;\n"
        "uniform vec4 LIGHTS_SPECULAR;\n"
        "varying vec2 texcoord;\n"
        "varying vec3 vVaryingLightDir;\n"
        "varying vec3 vVaryingNormal;\n"
        "void main(){\n"
        "    float maxVariance = 10.0;\n"
        "    float minVariance = maxVariance / 2.0;\n"
        "    vec3 normalAdjusted = vVaryingNormal + normalize(texture2D(NORMALMAP, texcoord.st).rgb * maxVariance - minVariance);\n"
        "    float diffuseIntensity = max(0.0, dot(normalize(normalAdjusted), normalize(vVaryingLightDir)));\n"
        "    vec3 colour = (diffuseIntensity * LIGHTS_DIFFUSE.rgb) * texture2D(TEXTURE0, texcoord.st).rgb + LIGHTS_AMBIENT.rgb;\n"
        "    vec4 vFragColour = vec4(colour, 1.0);\n"
        "    vec3 vReflection = normalize(reflect(-normalize(normalAdjusted), normalize(vVaryingLightDir)));\n"
        "    float specularIntensity = max(0.0, dot(normalize(normalAdjusted), vReflection));\n"
        "    if(diffuseIntensity > 0.98)\n"
        "    {\n"
        "        float fSpec = pow(specularIntensity, 64.0);\n"
        "        vFragColour.rgb += vec3(fSpec * LIGHTS_SPECULAR.rgb);\n"
        "    }\n"
        "    gl_FragColor = vFragColour;\n"
        "}";

struct rgb_t{
    rgb_t(uint8_t r = 0, uint8_t g = 0, uint8_t b = 0) : r(r), g(g), b(b) {}
    uint8_t r, g, b;
};

class RotationDelegate : public Ce::INodeDelegate{
public:
    virtual void onRenderNode(const std::shared_ptr<Ce::INode> &node, const Ce::SRenderState &renderState){
        node->rotate(glm::vec3(renderState.frameTime * 15.0f, renderState.frameTime * 15.0f, 0));
    }
};

class NormalMapDelegate : public Ce::INodeDelegate{
public:
    virtual void onRenderNode(const std::shared_ptr<Ce::INode> &node, const Ce::SRenderState &renderState){
        if (node->program())
            node->program()->setParameterTexture("NORMALMAP", node->material(0).textures[1], 1);
    }
};

std::shared_ptr<Ce::ICore> core;

void init(){
    core = Ce::createCore(Ce::createDriver(), glm::vec2(800, 480));
    auto program = core->driver()->createGpuProgram(simpleVs, simpleFs);

    auto texture = core->loadTexture("col.jpg");
    auto normalmap = core->loadTexture("col_n.jpg");

    auto scene = core->createScene();
    scene->registerProgram(program);
    scene->setClearColor(glm::vec4(0.5, 0.6, 0.7, 1.0));
    scene->setPerspective(45, core->windowSize().x / core->windowSize().y, 1.0, 100.0);
    scene->setViewMatrix(glm::lookAt(glm::vec3(0, 0, -5), glm::vec3(0, 0, 0), glm::vec3(0, 1, 0)));
    core->setScene(scene);

    auto node = core->createPreparedNode(Ce::E_PN_CUBE);
    node->setRotation(glm::vec3(25, 25, 0));
    node->setProgram(program);
    node->material(0).textures[0] = texture;
    node->material(0).textures[1] = normalmap;
    node->setParent(scene->rootNode());
    node->setNodeDelegate(std::shared_ptr<NormalMapDelegate>(new NormalMapDelegate));

    std::shared_ptr<Ce::INode> node2 = core->loadNode("dwarf.xml");
    //node2->setParent(scene->rootNode());
    node2->setPosition(glm::vec3(2, 0, 0));
    node2->setScale(glm::vec3(0.03, 0.03, 0.03));
    node2->setAnimation(0, 300, true, 10);

    Ce::SLightInfo lightInfo;
    lightInfo.ambient = glm::vec4(0.0, 0.0, 0.0, 1.0);
    lightInfo.diffuse = glm::vec4(1.0);
    auto light = core->createPreparedNode(Ce::E_PN_CUBE);
    auto lightDummy = core->createNode();
    light->enableLighting(lightInfo);
    lightDummy->setParent(scene->rootNode());
    light->setPosition(glm::vec3(0, 0, -2));
    lightDummy->setNodeDelegate(std::shared_ptr<RotationDelegate>(new RotationDelegate));
    lightDummy->addChild(light);
    light->setScale(glm::vec3(0.1, 0.1, 0.1));

    core->start();
}

int main_ce(){
    init();
    return 0;
}
